/* graph.c
 * Graph construction and manipulation
 */

#include "graph.h"

void construct_vertex_map (int *x, int *y, int m, int n, int asize, 
			   graph_t *graph)
{
  int i, j, k;

  graph->vmap = (int **) malloc_chk(m * n * sizeof (int *));
  for (i = 0; i < m * n; i++)
    graph->vmap[i] = (int *) malloc_chk(2 * sizeof (int));

  graph->revmap = (int **) malloc_chk(m * sizeof (int *));
  for (i = 0; i < m; i++) {
    graph->revmap[i] = (int *) malloc_chk(n * sizeof (int));
    for (j = 0; j < n; j++)
      graph->revmap[i][j] = -1;
  }

  graph->nv = 0;
  for (k = 0; k < asize; k++) {
    for (i = 0; i < m; i++)
      if (x[i] == k)
        for (j = 0; j < n; j++)
          if (y[j] == k) {
            graph->vmap[graph->nv][0] = i;
            graph->vmap[graph->nv][1] = j;
            graph->revmap[i][j] = graph->nv;
            (graph->nv)++;
          }
  }
}

void construct_graph (int m, int n, graph_t *graph)
{
  int i, j;

  graph->adj = (int **) malloc_chk((graph->nv + 2) * sizeof (int *));
  for (i = 0; i < graph->nv + 2; i++)
    graph->adj[i] = (int *) malloc_chk((graph->nv + 2) * sizeof (int));

  for (i = 0; i < m; i++)
    for (j = 0; j < n; j++) {
      if (graph->revmap[i][j] != -1) {
	printf("%d, %d\n", i, j);
	find_noncross_vertices(m, n, i, j, 1, graph->revmap[i][j], graph);
	find_cross_vertices(m, n, i, j, 1, graph->revmap[i][j], graph);
      }
    }

  /* graph->nv:   bottom left corner for noncross, top left corner for cross
   * graph->nv+1: top right corner for noncross, bottom right corner for cross
   */
  find_noncross_vertices(m, n, -1, -1, 0, graph->nv, graph);
  top_left_edges(m, n, graph->nv, graph);
  top_right_edges(m, n, graph->nv + 1, graph);
  find_cross_vertices_incl(m, n, m - 1, 0, 0, graph->nv + 1, graph);
}

void find_noncross_vertices (int m, int n, int i, int j, int reverse, 
			     int vij, graph_t *graph)
{
  int k, l;
  int nlimit, tmp;

  nlimit = n;
  tmp = 0;
  for (k = i + 1; k < m; k++) {
    for (l = j + 1; l < nlimit; l++)
      if (graph->revmap[k][l] != -1) {
	graph->adj[vij][graph->revmap[k][l]] = 1;
	if (reverse == 1)
	  graph->adj[graph->revmap[k][l]][vij] = 2;
	if (tmp == 0)
	  tmp = l + 1;
      }
    if (tmp != 0) {
      nlimit = tmp;
      tmp = 0;
    }
  }
}

void find_cross_vertices (int m, int n, int i, int j, int reverse, 
			  int vij, graph_t *graph)
{
  int k, l;
  int nlimit, tmp, jinit;

  nlimit = n;
  tmp = 0;
  jinit = j + 1;
  for (k = i; k >= 0; k--) {
    for (l = jinit; l < nlimit; l++)
      if (graph->revmap[k][l] != -1) {
	graph->adj[vij][graph->revmap[k][l]] = -1;
	if (reverse == 1)
	  graph->adj[graph->revmap[k][l]][vij] = -2;
	tmp = l;
	break;
      }
    if (tmp != 0) {
      nlimit = tmp;
      tmp = 0;
    }
    jinit = j;
  }
}

void find_cross_vertices_incl (int m, int n, int i, int j, int reverse, 
			       int vij, graph_t *graph)
{
  int k, l;
  int nlimit, tmp;

  nlimit = n;
  tmp = 0;
  for (k = i; k >= 0; k--) {
    for (l = j; l < nlimit; l++)
      if (graph->revmap[k][l] != -1) {
	graph->adj[vij][graph->revmap[k][l]] = -1;
	if (reverse == 1)
	  graph->adj[graph->revmap[k][l]][vij] = -2;
	tmp = l;
	break;
      }
    if (tmp != 0) {
      nlimit = tmp;
      tmp = 0;
    }
  }
}

void top_left_edges (int m, int n, int vij, graph_t *graph)
{
  int k, l;
  int nlimit, tmp;

  nlimit = 0;
  tmp = 0;
  for (k = 0; k < m; k++) {
    for (l = n - 1; l >= nlimit; l--)
      if (graph->revmap[k][l] != -1) {
	graph->adj[graph->revmap[k][l]][vij] = -1;
	tmp = l + 1;
	break;
      }
    if (tmp != 0) {
      nlimit = tmp;
      tmp = 0;
    }
  }
}

void top_right_edges (int m, int n, int vij, graph_t *graph)
{
  int k, l;
  int nlimit, tmp;

  nlimit = 0;
  tmp = 0;
  for (k = m - 1; k >= 0; k--) {
    for (l = n - 1; l >= nlimit; l--)
      if (graph->revmap[k][l] != -1) {
	graph->adj[graph->revmap[k][l]][vij] = 1;
	if (tmp == 0)
	  tmp = l;
      }
    if (tmp != 0) {
      nlimit = tmp;
      tmp = 0;
    }
  }
}

/* Not used. */
void construct_graph_complete (int *x, int *y, int m, int n, graph_t *graph)
{
  int i, j;

  graph->adj = (int **) malloc_chk(graph->nv * sizeof (int *));
  for (i = 0; i < graph->nv; i++)
    graph->adj[i] = (int *) malloc_chk(graph->nv * sizeof (int));

  for (i = 0; i < graph->nv; i++)
    for (j = 0; j < graph->nv; j++) {
      graph->adj[i][j] = 0;
      if (graph->vmap[j][0] > graph->vmap[i][0] && 
          graph->vmap[j][1] > graph->vmap[i][1]) {
        graph->adj[i][j] = 1;
        graph->adj[j][i] = 2;
      }
      else if (graph->vmap[j][0] <= graph->vmap[i][0] && 
               graph->vmap[j][1] >= graph->vmap[i][1]) {
        graph->adj[i][j] = -1;
        graph->adj[j][i] = -2;
      }
    }
}

void free_graph (int m, int n, graph_t *graph)
{
  int i;

  for (i = 0; i < m * n; i++)
    free(graph->vmap[i]);
  free(graph->vmap);

  for (i = 0; i < m; i++)
    free(graph->revmap[i]);
  free(graph->revmap);

  for (i = 0; i < graph->nv + 2; i++)
    free(graph->adj[i]);
  free(graph->adj);
}
