/* preprocess.c
 * Preprocessing functions for RFLCS
 */

#include "preprocess.h"

/* Preprocesses two sequences for solving RFLCS. Eliminates symbols that
 * only appear in one sequence and consecutive symbols, keeping one of them.
 * Also eliminates symbols of the form abababa..., keeping aba.
 */
int preprocess_rflcs (int *x, int *y, int *xr, int *yr, int m, int n,
		      int *mr, int *nr, int asize)
{
  int i, min, asizeel;
  int *xcount, *ycount;

  xcount = (int *) malloc_chk(asize * sizeof (int));
  ycount = (int *) malloc_chk(asize * sizeof (int));

  for (i = 0; i < asize; i++)
    xcount[i] = ycount[i] = 0;

  for (i = 0; i < m; i++)
    xcount[(unsigned int) x[i]]++;

  for (i = 0; i < n; i++)
    ycount[(unsigned int) y[i]]++;

  min = MIN(m, n);
  *mr = *nr = 0;
  for (i = 0; i < min; i++) {
    if (!((ycount[(unsigned int) x[i]] == 0 && 
           xcount[(unsigned int) x[i]] > 0) ||
          (*mr > 0 && x[i] == xr[*mr-1]) ||
          (*mr > 2 && x[i] == xr[*mr-2] && xr[*mr-1] == xr[*mr-3])))
      xr[(*mr)++] = x[i];
    if (!((xcount[(unsigned int) y[i]] == 0 && 
           ycount[(unsigned int) y[i]] > 0) ||
          (*nr > 0 && y[i] == yr[*nr-1]) ||
          (*nr > 2 && y[i] == yr[*nr-2] && yr[*nr-1] == yr[*nr-3])))
      yr[(*nr)++] = y[i];
  }

  if (min == n) {
    for (i = n; i < m; i++)
      if (!((*mr > 0 && x[i] == xr[*mr-1]) ||
            (*mr > 2 && x[i] == xr[*mr-2] && xr[*mr-1] == xr[*mr-3])))
        xr[(*mr)++] = x[i];
  }
  else {
    for (i = m; i < n; i++)
      if (!((*nr > 0 && y[i] == yr[*nr-1]) ||
            (*nr > 2 && y[i] == yr[*nr-2] && yr[*nr-1] == yr[*nr-3])))
        yr[(*nr)++] = y[i];
  }

  asizeel = asize;
  for (i = 0; i < asize; i++)
    if (xcount[i] == 0 || ycount[i] == 0)
      asizeel--;

  return asizeel;
}
