/* rflcs.c
 * Main file for RFLCS
 */

#include "rflcs.h"

int main (int argc, char *argv[])
{
  int i, k, m, n, asize, mpn;
  int metapost, genmethod, color, silent, integer, derand3, mgraph, approx4;
  int bgraph, load;
  int optc;
  int *x, *y, *z;
  char mpfname[MAX_FNAME], seqfname[MAX_FNAME];
  unsigned int seed;
  FILE *mpfile;
  graph_t *graph;

  static struct option long_options[] = {
    {"alphabet", 1, 0, 'a'},
    {"metapost", 1, 0, 'm'},
    {"generate", 1, 0, 'g'},
    {"color", 1, 0, 'c'},
    {"silent", 0, 0, 's'},
    {"seed", 1, 0, 1},
    {"integer", 0, 0, 'i'},
    {"derand3", 0, 0, 2},
    {"mgraph", 0, 0, 3},
    {"approx4", 0, 0, 4},
    {"buildgraph", 0, 0, 'b'},
    {"load", 1, 0, 'l'},
    {0, 0, 0, 0}
  };

  asize = DEFAULT_ASIZE;
  metapost = 0;
  genmethod = 1;
  color = 2;
  silent = 0;
  seed = 0;
  integer = 0;
  derand3 = 0;
  mgraph = 0;
  approx4 = 0;
  bgraph = 0;
  mpn = 1;
  load = 0;
  mpfile = NULL;
  graph = NULL;

  while (1) {
    optc = getopt_long(argc, argv, "a:m:g:c:sibl:", long_options, (int *) 0);
    if (optc == -1)
      break;

    switch (optc) {
      case 'a':
        asize = atoi_chk(optarg);
        break;
      case 'm':
        strcpy(mpfname, optarg);
        metapost = 1;
        break;
      case 'g':
        genmethod = atoi_chk(optarg);
        break;
      case 'c':
        color = atoi_chk(optarg);
	if (color < 0 || color > 2) {
          fprintf(stderr, "Error: Color type invalid.\n");
          exit(EXIT_FAILURE);
	}
        break;
      case 's':
        silent = 1;
        break;
      case 'i':
        integer = 1;
        break;
      case 'b':
        bgraph = 1;
        break;
      case 'l':
        strcpy(seqfname, optarg);
        load = 1;
        break;
      case 1: /* seed */
        seed = atoi_chk(optarg);
        break;
      case 2: /* derand3 */
        derand3 = 1;
        break;
      case 3: /* mgraph */
        mgraph = 1;
        break;
      case 4: /* approx4 */
        approx4 = 1;
        break;
      default:
        fprintf(stderr, "Error: getopt returned %c\n", optc);
        exit(EXIT_FAILURE);
    }
  }

  if (argc - optind < 2) {
    print_usage(argv[0]);
    exit(EXIT_FAILURE);
  }

  if (seed == 0)
    seed = time(NULL);

  if (silent == 0)
    printf("Seed = %d\n", seed);
  srand(seed);

  m = atoi_chk(argv[optind]);
  n = atoi_chk(argv[optind + 1]);

  if (metapost == 1) {
    mpfile = fopen(mpfname, "w");
    if (mpfile == NULL) {
      fprintf(stderr, "Error: Unable to open metapost file.\n");
      exit(EXIT_FAILURE);
    }
  }

  if (genmethod == 3) {
    x = (int *) malloc_chk(m * asize * sizeof (int));
    y = (int *) malloc_chk(n * asize * sizeof (int));
    z = (int *) malloc_chk(MIN(m * asize, n * asize) * sizeof (int));
  }
  else {
    x = (int *) malloc_chk(m * sizeof (int));
    y = (int *) malloc_chk(n * sizeof (int));
    z = (int *) malloc_chk(MIN(m, n) * sizeof (int));
  }

  if (load == 1)
    load_sequence(seqfname, x, y, &m, &n, &asize);
  else
    generate_sequence(x, y, &m, &n, asize, genmethod);

  if (silent == 0) {
    printf("x = ");
    print_sequence(x, m);
    printf("y = ");
    print_sequence(y, n);
    printf("Length of sequence x: %d\n", m);
    printf("Length of sequence y: %d\n", n);
    printf("Alphabet size: %d\n", asize);
    printf("\n");
  }

  if (bgraph == 1) {
    if (silent == 0)
      printf("Building graph... ");
    graph = (graph_t *) malloc_chk(sizeof (graph_t));
    construct_vertex_map(x, y, m, n, asize, graph);
    construct_graph(m, n, graph);
    if (silent == 0)
      printf("Done.\n");
  }

  k = run_algorithm(x, y, z, m, n, silent, metapost, mpfile, &mpn, 
                      color, asize, graph, 1);

  for (i = 2; i <= 4; i++)
    k = run_algorithm(x, y, z, m, n, silent, metapost, mpfile, &mpn, 
                      color, asize, graph, i);
  /* k is currently unused but left here in case needed, so at the end
   * of every iteration we have the solution stored in z with length k.
   */

  if (derand3 == 1)
    k = run_algorithm(x, y, z, m, n, silent, metapost, mpfile, &mpn, 
                      color, asize, graph, 5);

  if (approx4 == 1)
    k = run_algorithm(x, y, z, m, n, silent, metapost, mpfile, &mpn, 
                      color, asize, graph, 6);

  if (integer == 1) {
    k = run_algorithm(x, y, z, m, n, silent, metapost, mpfile, &mpn, 
                      color, asize, graph, 7);
  }
  else {
    k = run_algorithm(x, y, z, m, n, silent, metapost, mpfile, &mpn, 
		      color, asize, graph, 8);
  }

  if (metapost == 1 && mgraph == 1) {
    if (bgraph == 1) {
      print_metapost_graph_done(x, y, m, n, mpfile, mpn++, color, asize, 0,
                                graph);
      print_metapost_graph_done(x, y, m, n, mpfile, mpn++, color, asize, 1,
                                graph);
    }
    else {
      print_metapost_graph(x, y, m, n, mpfile, mpn++, color, asize, 0, 0);
      print_metapost_graph(x, y, m, n, mpfile, mpn++, color, asize, 0, 1);
      print_metapost_graph(x, y, m, n, mpfile, mpn++, color, asize, 1, 0);
      print_metapost_graph(x, y, m, n, mpfile, mpn++, color, asize, 1, 1);
    }
  }
  
  if (metapost == 1) {
    fprintf(mpfile, "end;\n");
    fclose(mpfile);
  }

  if (bgraph == 1)
    free_graph(m, n, graph);

  free(x);
  free(y);
  free(z);

  return 0;
}

/* A wrapper to run an algorithm according to alg.
 */
int run_algorithm (int *x, int *y, int *z, int m, int n, int silent, 
                   int metapost, FILE *mpfile, int *mpn, int color, 
                   int asize, graph_t *graph, int alg)
{
  int k;
  struct tms t_start, t_end;
  double time_elapsed;
#ifdef MEAN_SILENCE
  int fd;
#endif

  if (silent == 0) {
    switch (alg) {
      case 1:
        printf("With repetitions: \n");
        break;
      case 2:
        printf("Approximation 1: \n");
        break;
      case 3:
        printf("Approximation 2: \n");
        break;
      case 4:
        printf("Approximation 3: \n");
        break;
      case 5:
        printf("Approximation 3 (derandomized): \n");
        break;
      case 6:
        printf("Approximation 4 (simplex-based): \n");
        break;
      case 7:
        printf("Integer programming: \n");
        break;
      case 8:
        printf("Branch-and-cut based solver: \n");
        break;
      default:
        fprintf(stderr, "Error: Algorithm identifier not recognized.\n");
        exit(EXIT_FAILURE);     
    }
  }

  times(&t_start);

#ifdef MEAN_SILENCE
  if (silent == 1) {
    fd = dup(STDOUT_FILENO);
    freopen("/dev/null", "w", stdout);
  }
#endif

  switch (alg) {
    case 1:
      k = lcs_cormen(x, y, z, m, n);
      break;
    case 2:
      k = rflcs_approx1(x, y, z, m, n, asize);
      break;
    case 3:
      k = rflcs_approx2(x, y, z, m, n, 1, asize);
      break;
    case 4:
      k = rflcs_approx2(x, y, z, m, n, 2, asize);
      break;
    case 5:
      k = rflcs_approx2_derand(x, y, z, m, n, asize);
      break;
    case 6:
      k = rflcs_approx2(x, y, z, m, n, 3, asize);
      break;
    case 7:
      k = rflcs_ip(x, y, z, m, n, asize, graph);
      break;
    case 8:
      k = rflcs_bc(x, y, z, m, n, asize);
      break;
  }
  times(&t_end);

#ifdef MEAN_SILENCE
  if (silent == 1) {
    fflush(stdout);
    dup2(fd, STDOUT_FILENO);
  }
#endif

  time_elapsed = (((t_end.tms_utime + t_end.tms_cutime) -
		  (t_start.tms_utime + t_start.tms_cutime)) /
                  sysconf(_SC_CLK_TCK));

  if (silent == 0) {
    printf("Solution: ");
    print_sequence(z, k);
    printf("Length: %d\n", k);
    printf("Time elapsed: %f\n\n", time_elapsed);
  }
  else
    printf("%d\n", k);

  if (metapost == 1)
    print_metapost(x, y, z, m, n, k, mpfile, (*mpn)++, color, asize);

  return k;
}
